#include <stdlib.h>
#include <stdio.h>
#include "error.h"
#include "singlio.h"
#include "protos.h"
#include "getopt.h"
#include "cmdline.h"
#include "physics.h"
#include "mpmy.h"
#include "gc.h"

#define NOBJ 1024
#define LGPGSIZE 14
#define WSETPAGES 256
#define ERRTOL 0.05F
#define EPSILON 0.01F
#define RCRITFAC 4.0F /* see comment at top of dl.c */
#define BKTSPACE (16*1024*1024)
#define NBUCKETS 0    /* use bktspace by default */
#define DT 0.01
#define NSTEPS 1
#define OCCUPANCY 3
#define MAX_SINKSPLIT 1
#define TIMEOUT 0

#define OPT_NOBJ 1001
#define OPT_LGPGSIZE 1002
#define OPT_WSETPAGES 1004
#define OPT_MSG 1005
#define OPT_EPSILON 1006
#define OPT_ERRTOL 1007
#define OPT_NBUCKETS 1008
#define OPT_DT 1009
#define OPT_NSTEPS 1010
#define OPT_DATA 1011
#define OPT_SWAP 1012
#define OPT_HASH_CHECK 1013
#define OPT_CD 1014
#define OPT_OCCUPANCY 1015
#define OPT_DECOMPTOL 1016
#define OPT_RCRITFAC 1017
#define OPT_MAX_SINKSPLIT 1018
#define OPT_TIMEOUT  1019
#define OPT_BKTSPACE 1020

/* These are only exposed in 'cmdline.h', which main will include, but
   nobody else */
int lgpgsize = LGPGSIZE;
int wsetpages = WSETPAGES;
int bktspace = BKTSPACE;
int nbuckets = NBUCKETS;
char *datafile = NULL;
char *swapfile_fmt = "swapfile.%d";
char swapfile[512];
char *newdir_fmt = "misc";
char newdir[512];
int nsteps = NSTEPS;
int do_print_tree;
int do_nlgn = 0;
int do_dl = 0;
int do_vectorize = 0;
int do_check_tree = 0;
char *msg_arg = NULL;
float decomptol = 0.1;		/* i.e., 10% of sum of wgts of nbr procs*/
int max_sinksplit_per_body = MAX_SINKSPLIT;
int timeout = TIMEOUT;

/* These are 'public' and visible in physics.h too */
int verbose = 0;
int debug = 0;
int gnobj = NOBJ;
float errtol = ERRTOL;
float epsilon = EPSILON;
float rcritfac = RCRITFAC;
float dt = DT;
int key_order = PH_ORDER;
int occupancy = OCCUPANCY;

char *optstring = "";

struct option optab[] ={
    {"nobj", required_argument, NULL, OPT_NOBJ},
    {"datafile", required_argument, NULL, OPT_DATA},
    {"swapfile", required_argument, NULL, OPT_SWAP},
    {"cd", required_argument, NULL, OPT_CD},
    {"lgpgsize", required_argument, NULL, OPT_LGPGSIZE},
    {"bktspace", required_argument, NULL, OPT_BKTSPACE},
    {"nbuckets", required_argument, NULL, OPT_NBUCKETS},
    {"wsetpages", required_argument, NULL, OPT_WSETPAGES},
    {"msgs", required_argument, NULL, OPT_MSG},
    {"errtol", required_argument, NULL, OPT_ERRTOL},
    {"epsilon", required_argument, NULL, OPT_EPSILON},
    {"rcritfac", required_argument, NULL, OPT_RCRITFAC},
    {"max_sinksplit_per_body", required_argument, NULL, OPT_MAX_SINKSPLIT},
    {"dt", required_argument, NULL, OPT_DT},
    {"nsteps", required_argument, NULL, OPT_NSTEPS},
    {"printtree", no_argument, &do_print_tree, 1},
    {"dl", no_argument, &do_dl, 1},
    {"nlgn", no_argument, &do_nlgn, 1},
    {"check", no_argument, &do_check_tree, 1},
    {"vectorize", no_argument, &do_vectorize, 1},
    {"novectorize", no_argument, &do_vectorize, 0},
    {"nodl", no_argument, &do_dl, 0},
    {"nonlgn", no_argument, &do_nlgn, 0},
    {"nocheck", no_argument, &do_check_tree, 0},
    {"morton", no_argument, &key_order, MORTON_ORDER},
    {"peano", no_argument, &key_order, PH_ORDER},
    {"occupancy", required_argument, NULL, OPT_OCCUPANCY},
    {"verbose", no_argument, &verbose, 1},
    {"decomptol", required_argument, NULL, OPT_DECOMPTOL},
    {"timeout", required_argument, NULL, OPT_TIMEOUT},
    {"debug", no_argument, &debug, 1},
    { 0, 0, 0, 0}
};

void
ParseArgs(int argc, char **argv){
  int c;
  char *endarg;

  for(optopt=0;
      (c=getopt_long_only(argc, argv, optstring, optab, NULL)) != EOF;
      optopt=0){
    switch(c){
    case OPT_NOBJ:
      gnobj = atoi(optarg);
      break;
    case OPT_LGPGSIZE:
      lgpgsize = atoi(optarg);
      break;
    case OPT_WSETPAGES:
      wsetpages = atoi(optarg);
      break;
    case OPT_MSG:
      msg_arg = optarg;
      break;
    case OPT_ERRTOL:
      errtol = atof(optarg);
      break;
    case OPT_EPSILON:
      epsilon = atof(optarg);
      break;
    case OPT_DT:
      dt = atof(optarg);
      break;
    case OPT_RCRITFAC:
      rcritfac = atof(optarg);
      break;
    case OPT_MAX_SINKSPLIT:
      max_sinksplit_per_body = atoi(optarg);
      break;
    case OPT_NBUCKETS:
      nbuckets = atoi(optarg);
      break;
    case OPT_NSTEPS:
      nsteps = atoi(optarg);
      break;
    case OPT_DATA:
      datafile = optarg;
      break;
    case OPT_SWAP:
      swapfile_fmt = optarg;
      break;
    case OPT_CD:
      newdir_fmt = optarg;
      break;
    case OPT_OCCUPANCY:
      /* The terminal 'm' parameter */
      occupancy = atoi(optarg);
      break;
    case OPT_DECOMPTOL:
      decomptol = atof(optarg);
      break;
    case OPT_TIMEOUT:
      timeout = atoi(optarg);
      break;
    case OPT_BKTSPACE:
      bktspace = strtoul(optarg, &endarg, 0);
      if( *endarg == 'm' || *endarg == 'M' )
	bktspace *= 1024*1024;
      if( *endarg == 'k' || *endarg == 'K' )
	bktspace *= 1024;
      break;
    case 0:
      /* A long option that got handled with an assignment */
      break;
    }
  }

  sprintf(swapfile, swapfile_fmt, MPMY_Procnum());
  singlPrintf("Build tree wtih gnobj=%d, m=%d\n", 
	      gnobj, occupancy);
  singlPrintf("Swapping to %s, pgsize=%d, wset=%d pages\n", 
	      swapfile_fmt, 1<<lgpgsize, wsetpages);
  if( nbuckets == 0 ){
    nbuckets = 1.5 * ((double)gnobj * (double)sizeof(body))/(MPMY_Nproc()*bktspace) ;
    if( nbuckets <= 0 )
      nbuckets = 1;
    singlPrintf("target bucket footprint: %.3fMB\n", bktspace/(1024.*1024.));
  }
  singlPrintf("nbuckets=%d\n", nbuckets);
  singlPrintf("errtol=%g, epsilon=%g, dt=%g, key_order=%s, decomp_tol=%g\n",
	      errtol, epsilon, dt, key_order==MORTON_ORDER?"morton":"peano",
	      decomptol);
  if(do_dl){
    singlPrintf("rcritfac=%g, max %d sinksplits per body\n",
		rcritfac, max_sinksplit_per_body);
  }
  /* Should we do sanity checkes here or elsewhere? */
  if( decomptol >= 1.0 || decomptol < 0. ){
    Error("--decomptol %g: arg out of range.  Must be in [0.0, 1.0)\n", decomptol);
  }
}

